/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.jsvgtest;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.CubicCurve2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.awt.geom.QuadCurve2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.io.File;

import javax.swing.JPanel;

import com.jsevy.jsvg.*;





public class SVGTestPanel extends JPanel
{
        
    private static final long serialVersionUID = 1L;

    // what to display and generate in the PDF file - just use button ID
    //private int displayContent = R.id.testGraphicsButton;
    private String displayContent = "graphics";
    
    //private float displayScale = 1.0f;
    
    private int xCen = 0;
    private int yCen = 0;
    private int height;
    private int width;
    
    private String fontFilePath = "";
    
    SVGDocument svgDocument = new SVGDocument();
    
    // create a random transformation
    private double randomRotation = 90 * 2*(Math.random() - 0.5);
    private double randomShearX = Math.random() - 0.5;
    private double randomShearY = Math.random() - 0.5;
    private double randomScaleX = Math.random() + 0.5;
    private double randomScaleY = Math.random() + 0.5;
    private double randomStartAngle = 180 * (2 * Math.random() - 0.5);
    private double randomSweepAngle = 360 * (2 * Math.random() - 0.5);
    private Color randomColor = Color.BLUE;
    
    // little utility class
    public class RealPoint
    {
        double x;
        double y;
        double z;
        
        public RealPoint(double x, double y, double z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
    }
    
    
    
    public void setDisplayContent(String content)
    {
        this.displayContent = content;
        
        if (content.equals("random"))
        {
            randomRotation = 90 * 2*(Math.random() - 0.5);
            randomShearX = Math.random() - 0.5;
            randomShearY = Math.random() - 0.5;
            randomScaleX = Math.random() + 0.5;
            randomScaleY = Math.random() + 0.5;
            randomStartAngle = 180 * (2 * Math.random() - 0.5);
            randomSweepAngle = 360 * (2 * Math.random() - 0.5);
            randomColor = new Color((float)Math.random(), (float)Math.random(), (float)Math.random());
        }
        
        repaint();
    }
    
    
    public void setFontFilePath(String fontFilePath)
    {
        this.fontFilePath = fontFilePath;
        repaint();
    }
    
    
    
    public SVGTestPanel()
    {
        xCen = 1000;
        yCen = 2000;
        height = 5000;
        width = 5000;
        
        setPreferredSize(new Dimension(width, height));
    }
    
    
    
    public void paint(Graphics g)
    {
        g.clearRect(0, 0, width, height);
        g.translate(xCen, yCen);
        drawDesign((Graphics2D)g);
    }
    
    
    
    public String generateSVG()
    {
        svgDocument = new SVGDocument();
        
        // set units to cm
        svgDocument.setUnits(SVGDocument.Units.CENTIMETERS);
        
        // get the ODGGraphics object to draw into
        SVGGraphics svgGraphics = svgDocument.getGraphics();
        
        // supply this as a Graphics subclass object to our main Draw method
        drawDesign(svgGraphics);
        
        // set the document size to enclose all graphics elements - now happens automatically,
        // so no need to do this any more
        //Rectangle2D.Double bounds = svgDocument.getEntityBounds();
        //System.out.println("bounds: " + bounds.x + ", " + bounds.y + ", " + bounds.width + ", " + bounds.height);
        //svgDocument.setDocumentBounds(bounds);
        
        // get the ODG output (just XML text)
        String stringOutput = new String();
        try
        {
            stringOutput = svgDocument.toSVGString();
        }
        catch(Exception e)
        {
            stringOutput = e.getMessage();
        }
        
        System.out.println(stringOutput);
        
        return stringOutput;
    }
    
    
    
    /**
     * Used to draw both on the screen and to the PDF document, to check if the PDF image matches the
     * one on the screen
     * @param canvas  Either a "normal" canvas for screen rendering or one associated with a PDF Page
     */
    private void drawDesign(Graphics2D graphics)
    {
        
        
        // fill the page with a pale yellow
        //graphics.setColor(new Color(255, 245, 240));
        //graphics.clearRect(0, 0, this.getWidth(), this.getHeight());
        
        if (displayContent.equals("graphics"))
        {
            testGraphics(graphics);
        }
        
        if (displayContent.equals("text"))
        {
            testText(graphics);
        }
        
        if (displayContent.equals("random"))
        {
            testRandom(graphics);
        }
        
        if (displayContent.equals("transformations_graphics"))
        {
            testTransformationsGraphics(graphics);
        }
        
        if (displayContent.equals("transformations_text"))
        {
            testTransformationsText(graphics);
        }
        
        if (displayContent.equals("graphics2D"))
        {
            testGraphics2D(graphics);
        }
        
        if (displayContent.equals("fill"))
        {
            testFill(graphics);
        }
        
        if (displayContent.equals("empty"))
        {
            testEmpty(graphics);
        }
        
        if (displayContent.equals("animation"))
        {
            testAnimation(graphics);
        }
           
        
    }
    
    
    
    private void testEmpty(Graphics graphics)
    {
        
    }
    
    
    
    private void testAnimation(Graphics2D graphics)
    {
        
        // draw a rectangle, and change its color twice with an SVG <set>; note that we can supply
        // either the SVG-standard color format string "#rrggbb", or a Java color, which will be
        // converted to the corresponding SVG color string. Note that we change the opacity for the
        // duration of these changes since the shape was not originally filled and thus has initial opacity 0
        graphics.setColor(Color.BLUE);
        graphics.setStroke(new BasicStroke(3));
        graphics.drawRect(1000, 500, 150, 150);
        svgDocument.setAttribute("fill-opacity", 1, 2 , 5, false);
        svgDocument.setAttribute("fill", "#ff00ff", 2 , 3, false);
        svgDocument.setAttribute("fill", Color.ORANGE, 3 , 5, false);
        
        
        // do an oval that pulses and changes color; use "true" as the last argument to scale
        // the radius value from Java values to SVG values
        graphics.setColor(Color.GREEN);
        graphics.setStroke(new BasicStroke(2));
        graphics.drawOval(1000, 1000, 400, 200);
        svgDocument.animateAttribute("rx", 200, 100, 1.500 , 3, 3, true, true);
        svgDocument.setAttribute("fill-opacity", 1, 3 , Double.MAX_VALUE, false);
        svgDocument.setAttribute("fill", Color.YELLOW, 3 , Double.POSITIVE_INFINITY, false);
        
        
        // try a rectangle that animates through several sizes; we supply a semicolon-separated list
        // of values for the animation sequence, and "true" for the last argument to scale these values
        graphics.setColor(Color.ORANGE);
        graphics.setStroke(new BasicStroke(2));
        graphics.drawRect(500, 1000, 400, 200);
        svgDocument.animateAttribute("width", "400;300;100;400", 4 , 2, Integer.MAX_VALUE, true, true);
        
        
        // do a circle that shrinks 3 times from its initial radius of 75 to 50; the fromValue
        // argument of "" indicates to start with the initial radius. Also animate the fill
        // color, going from red to yellow; since it starts filled, no need to set the fill opacity
        graphics.setColor(Color.RED);
        graphics.fillOval(1500, 500, 150, 150);
        svgDocument.animateAttribute("r", "", 50, 0.500 , 2, 3, true, true);
        svgDocument.animateAttribute("fill", "#ff0000;#ffff00;#ff0000", 4 , 2, Integer.MAX_VALUE, true, true);
        
        /*
        // get the SVG rectangle so we can add a SVGSet to it; alternative to the simplified approach
        // in case for some reason we want to apply other sets or animates later
        // if we're in Java with standard Graphics, getLastEntity() will return null, but this
        // is handled gracefully by the setAttribute method (does nothing if entity is null)
        SVGEntity svgEntity = svgDocument.getLastEntity();
        svgDocument.setAttribute(svgEntity, "fill", "#ff00ff", 2 , 3, false);
        svgDocument.setAttribute(svgEntity, "fill-opacity", 1, 2 , 3, false);
        
        // line to corner of rectangle
        graphics.drawLine(0, 0, 1000, 500);
        
        
        // do a circle that shrinks
        graphics.setColor(Color.RED);
        graphics.setStroke(new BasicStroke(3));
        graphics.drawOval(1500, 500, 150, 150);
        
        svgEntity = svgDocument.getLastEntity();
        svgDocument.animateAttribute(svgEntity, "r", "", 50, 0.500 , 2, 3, true, true);
        */
        
        // automated document sizing doesn't take into account size changes due to animations and sets
        svgDocument.setDocumentSize(2000, 2000);
    }
    
    
    

    private void testText(Graphics graphics)
    {
        // do some stuff
        FontMetrics fm;
        int ascent, descent, width, baseX, baseY;
        String text;
        
        // use the FontMetrics to draw a box around the text. Unfortunately, since the ODG fonts will in general be different
        // from the corresponding Java fonts, the boxes will be the wrong size in the ODG display.
        graphics.setFont(new Font(Font.SERIF, Font.PLAIN, 24));
        graphics.setColor(Color.RED);
        
        baseX = 150;
        baseY = 250;
        text = "Some 24-point serif red text at position " + baseX + ", " + baseY + " with some accented characters: áéíóúüèìç";
        graphics.drawLine(0, 0, baseX, baseY);
        graphics.drawString(text, baseX, baseY);
        fm = graphics.getFontMetrics();
        ascent = fm.getAscent();
        descent = fm.getDescent();
        width = fm.stringWidth(text);
        graphics.drawLine(baseX, baseY, baseX + width, baseY);
        graphics.drawRect(baseX, baseY - ascent, width, ascent + descent);
        
        graphics.setFont(new Font(Font.SANS_SERIF, Font.BOLD, 48));
        graphics.setColor(Color.GREEN);
        
        baseX = 180;
        baseY = 350;
        text = "Some 48-point sans-serif green text at position " + baseX + ", " + baseY;
        graphics.drawLine(0, 0, baseX, baseY);
        graphics.drawString(text, baseX, baseY);
        fm = graphics.getFontMetrics();
        ascent = fm.getAscent();
        descent = fm.getDescent();
        width = fm.stringWidth(text);
        graphics.drawLine(baseX, baseY, baseX + width, baseY);
        graphics.drawRect(baseX, baseY - ascent, width, ascent + descent);
        
        graphics.setFont(new Font(Font.MONOSPACED, Font.ITALIC, 38));
        graphics.setColor(Color.BLUE);
        
        baseX = 480;
        baseY = 400;
        text = "Some 38-point italic monospace blue text at position " + baseX + ", " + baseY;
        graphics.drawLine(0, 0, baseX, baseY);
        graphics.drawString(text, baseX, baseY);
        fm = graphics.getFontMetrics();
        ascent = fm.getAscent();
        descent = fm.getDescent();
        width = fm.stringWidth(text);
        graphics.drawLine(baseX, baseY, baseX + width, baseY);
        graphics.drawRect(baseX, baseY - ascent, width, ascent + descent);
        
        graphics.setFont(new Font(Font.SANS_SERIF, Font.BOLD|Font.ITALIC, 18));
        graphics.setColor(Color.BLACK);
        
        baseX = 580;
        baseY = 600;
        text = "Some 18-point bold-italic sans-serif black text at position " + baseX + ", " + baseY;
        graphics.drawLine(0, 0, baseX, baseY);
        graphics.drawString(text, baseX, baseY);
        fm = graphics.getFontMetrics();
        ascent = fm.getAscent();
        descent = fm.getDescent();
        width = fm.stringWidth(text);
        graphics.drawLine(baseX, baseY, baseX + width, baseY);
        graphics.drawRect(baseX, baseY - ascent, width, ascent + descent);
        
        
        // try drawing characters at indicated positions
        /*
        paint.setTypeface(Typeface.create(Typeface.SERIF, Typeface.ITALIC));
        float[] pos = { 400, 500, 440, 510, 480, 525, 520, 545, 560, 570, 600, 600, 640, 635, 680, 675, 720, 720, 760, 770, 800, 825};
        String text = "drawPosText";
        graphics.drawPosText(text, pos);
        */
        
        // draw characters along path
        /*
        paint.setTypeface(Typeface.create(Typeface.SERIF, Typeface.BOLD));
        Path path = new Path();
        path.addArc(new RectF(100, 1000, 700, 1400), 180, -180);
        graphics.drawArc(new RectF(100, 1000, 700, 1400), 180, -180, false);
        path.arcTo(new RectF(700, 1000, 1400, 1400), 180, 180);
        graphics.drawArc(new RectF(700, 1000, 1400, 1400), 180, 180, false);
        path.arcTo(new RectF(1400, 1100, 1600, 1300), 180, -180);
        graphics.drawArc(new RectF(1400, 1100, 1600, 1300), 180, -180, false);
        text = "Example of bold serif drawTextOnPath with offset and composite path";
        graphics.drawTextOnPath(text, path, 100, 40);
        */
        
    }
    
    
    /*
    private void testFonts(Graphics graphics)
    {
        int baseX, baseY;
        
        // load the Font from a path previously specified through a Dialog
        File fontFile = new File(fontFilePath);
        
        try 
        {
            Font newFont = Font.createFont(Font.TRUETYPE_FONT, fontFile);
            
            GraphicsEnvironment graphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
            graphicsEnvironment.registerFont(newFont);
            
            // Draw some strings inside bounding boxes. Since we're loading the same TrueType font into both Java and the CAD
            // program, one would expect the strings to fit inside the boxes the same in both cases; but alas, it's just 
            // an approximate fit, especially regarding string width. Not sure why this should be, but it is...
            graphics.setFont(newFont.deriveFont(24.0f));
            graphics.setColor(Color.RED);
            baseX = 150;
            baseY = 250;
            drawFontStringInBox(graphics, baseX, baseY);
            
            
            // try different sizes
            graphics.setFont(newFont.deriveFont(48.0f));
            graphics.setColor(Color.GREEN);
            baseX = 180;
            baseY = 350;
            drawFontStringInBox(graphics, baseX, baseY);
            
            
            graphics.setFont(newFont.deriveFont(38.0f));
            graphics.setColor(Color.BLUE);
            baseX = 480;
            baseY = 400;
            drawFontStringInBox(graphics, baseX, baseY);
        
        }
        catch (Exception e)
        {
            System.out.println("Problem loading font: " + e.toString());
        }
        
    }
    
    
    private void drawFontStringInBox(Graphics graphics, int baseX, int baseY)
    {
        FontMetrics fm = graphics.getFontMetrics();
        int capHeight = (int)Math.ceil(graphics.getFont().createGlyphVector(((Graphics2D) graphics).getFontRenderContext(),"H").getVisualBounds().getHeight());
        Font font = graphics.getFont();
        
        String text = "Some " + font.getSize() + "-point text in font " + font.getFamily() + " at position " + baseX + ", " + baseY + ".";
        int width = fm.stringWidth(text);
        
        graphics.drawLine(0, 0, baseX, baseY);
        graphics.drawString(text, baseX, baseY);
        graphics.drawRect(baseX, baseY - capHeight, width, capHeight);
        
        //System.out.println("Height, ascent, descent, leading: " + height + ", " + ascent + ", " + descent + ", " + leading);
    }
    
    */
    
    
    
    
    
    private void testGraphics(Graphics2D graphics)
    {
        // do some stuff
        
        // lines
        graphics.drawLine(0, 0, 200, 100);
        
        graphics.setColor(Color.GREEN);
        graphics.setStroke(new BasicStroke(5));
        graphics.drawLine(100, 200, 0, 100);
        
        graphics.setColor(Color.RED);
        graphics.setStroke(new BasicStroke(3));
        graphics.drawLine(0, 100, 100, -100);
        
        graphics.setColor(Color.YELLOW);
        graphics.setStroke(new BasicStroke(7));
        graphics.drawLine(0, 0, 800, 800);
        
        
        graphics.setColor(Color.BLACK);
        graphics.setStroke(new BasicStroke(1));
        
        // circles, ellipses and arcs
        
     // test line types
        float[] dashArray;
        BasicStroke stroke;
        
        int radius = 20;
        int center;
        
        graphics.drawLine(0, 0, 800, 800);
        
        dashArray = new float[2];
        dashArray[0] = 1; 
        dashArray[1] = 3;
        stroke = new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 1, dashArray, 0);
        graphics.setStroke(stroke);
        graphics.drawOval(715, 715, 170, 170);
        
        dashArray = new float[4];
        dashArray[0] = 1; 
        dashArray[1] = 3;
        dashArray[2] = 1; 
        dashArray[3] = 8;
        stroke = new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 1, dashArray, 0);
        graphics.setStroke(stroke);
        graphics.drawOval(790, 790, 20, 20);
        
        dashArray = new float[4];
        dashArray[0] = 1; 
        dashArray[1] = 3;
        dashArray[2] = 4; 
        dashArray[3] = 8;
        stroke = new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 1, dashArray, 0);
        graphics.setStroke(stroke);
        radius = 50;
        center = 800;
        graphics.drawArc(center-radius, center-radius, 2*radius, 2*radius, 0, 45);
        radius = 100;
        graphics.drawArc(center-radius, center-radius, 2*radius, 2*radius, 0, 90);
        radius = 150;
        graphics.drawArc(center-radius, center-radius, 2*radius, 2*radius, 0, -135);
        
        

        
        
        // draw a line to the center of the circles
        graphics.drawLine(0, 0, center, center);
        
        
        // do some in red, with a red thick outline
        graphics.setColor(Color.RED);
        graphics.setStroke(new BasicStroke(3));
        radius = 50;
        center = 400;
        graphics.drawArc(center-radius, center-radius, 2*radius, 2*radius, 90, 45);
        radius = 100;
        graphics.drawArc(center-radius, center-radius, 2*radius, 2*radius, 90, 90);
        radius = 150;
        graphics.drawArc(center-radius, center-radius, 2*radius, 2*radius, 90, -135);
        
        
        // draw a rectangle
        graphics.setColor(Color.BLUE);
        graphics.setStroke(new BasicStroke(3));
        graphics.drawRect(1000, 500, 150, 150);
        
        // line to corner of rectangle
        graphics.drawLine(0, 0, 1000, 500);
        
        
        // draw nested rounded rectangles; inner one has circular corners,
        // so will be represented as an ODGRectangle, while outer one has
        // elliptical corners, so will be represented as an ODGPath
        graphics.drawRoundRect(20, 200, 130, 100, 20, 10);
        graphics.drawRoundRect(30, 210, 110, 80, 30, 30);
        
        // draw line to where corner of outer rectangle would be if not for the rounded corner
        graphics.drawLine(0, 0, 20, 200);
        
        
        // draw some ellipses and elliptical arcs, contained in rectangles
        graphics.setColor(Color.MAGENTA);
        graphics.setStroke(new BasicStroke(2));
        graphics.drawLine(0, 0, 300, 1000);
        graphics.drawOval(200, 800, 200, 400);
        graphics.drawRect(200, 800, 200, 400);
        
        // draw oval arc
        graphics.setStroke(new BasicStroke(1));
        graphics.drawLine(0, 0, 300, 2000);
        graphics.drawArc(100, 1900, 400, 200, 60, 150);
        graphics.drawRect(100, 1900, 400, 200);
        
        // draw some paths
        Path2D.Double path1 = new Path2D.Double(Path2D.WIND_EVEN_ODD);
        
        graphics.setColor(Color.BLACK);
        graphics.setStroke(new BasicStroke(1));
        
        // Draw square at 1, 1 calling closePath() to draw last segment of square
        path1.moveTo(500.0, 100.0);
        path1.lineTo(600.0, 100.0);
        path1.lineTo(600.0, 200.0);
        path1.lineTo(500.0, 200.0);
        path1.closePath();
        
        
        // Add another square in center of larger square
        path1.append(new Rectangle.Double(525, 125, 50, 50), false);
        
        graphics.draw(path1);
        
       
        // draw a line to the corner of the rectangular ring
        graphics.drawLine(0, 0, 500, 100);
        
        
        // polyline and polygon
        int xPoints[] = {500, 600, 700, 800};
        int yPoints[] = {300, 400, 200, 300};
        graphics.drawPolyline(xPoints, yPoints, 4);
        
        int x2Points[] = {600, 700, 800, 700};
        int y2Points[] = {500, 400, 500, 600};
        graphics.drawPolygon(x2Points, y2Points, 4);
        
        
        
    }
    
    
    
    private void testFill(Graphics2D graphics)
    {
        graphics.setColor(Color.YELLOW);
        graphics.fillOval(250, 100, 100, 100);
        
        // get fancy - even apply a transformation
        // save current transform to restore it
        AffineTransform initialTransform = graphics.getTransform();
        
        graphics.rotate(-15 *(Math.PI/180)); // counterclockwise...
        graphics.shear(0.1f, 0.2f); //--> shear/skew known to be broken in Libreoffice Draw
        
        
        graphics.fillRect(100, 100, 100, 50);
        
        graphics.setColor(Color.CYAN);
        graphics.fillOval(250, 100, 100, 50);
        
        graphics.setColor(Color.RED);
        int[] xPoints = {200, 300, 150};
        int[] yPoints = {300, 350, 400};
        graphics.fillPolygon(xPoints , yPoints , xPoints.length);
        graphics.drawLine(0, 0, 200, 300);
        
        
        graphics.setColor(Color.GREEN);
        graphics.fillRoundRect(100, 400, 100, 150, 40, 60);
        graphics.drawLine(0, 0, 100, 400);
        
        
        
        graphics.setColor(Color.BLUE);
        graphics.fillArc(50, 600, 200, 100, 120, 120);
        graphics.setColor(Color.RED);
        graphics.drawArc(50, 600, 200, 100, 120, 120);
        graphics.setColor(Color.CYAN);
        graphics.fillArc(300, 600, 400, 200, 30, -60);
        graphics.setColor(Color.YELLOW);
        graphics.fillArc(100, 700, 300, 300, 0, 270);
        graphics.setColor(Color.GREEN);
        graphics.fillArc(100, 600, 200, 200, 30, 60);
        
        
        // fill arcs with different end treatments
        
        graphics.setTransform(initialTransform);
        graphics.setColor(Color.BLUE);
        
        Arc2D.Double arc2D = new Arc2D.Double(600, -100, 100, 200, -30, 120, Arc2D.OPEN);
        graphics.fill(arc2D);
        arc2D = new Arc2D.Double(800, -100, 100, 200, -30, 120, Arc2D.CHORD);
        graphics.fill(arc2D);
        arc2D = new Arc2D.Double(1000, -100, 100, 200, -30, 120, Arc2D.PIE);
        graphics.fill(arc2D);
        
        /*
        
        // now create a Path for use with the general draw(Shape) method
        // do as shape with hole using two disjoint paths; add winding specification so hole won't be filled
        graphics.setColor(Color.BLACK);
        Path2D.Double path1 = new Path2D.Double(Path2D.WIND_EVEN_ODD);
        
        // Draw square at 1, 1 calling closePath() to draw last segment of square
        path1.moveTo(500.0, 100.0);
        path1.lineTo(600.0, 100.0);
        path1.lineTo(600.0, 200.0);
        path1.lineTo(500.0, 200.0);
        path1.closePath();
        
        // Add another square in center of larger square
        path1.append(new Rectangle.Double(525, 125, 50, 50), false);
        
        graphics.fill(path1);
        
        // draw a line to the corner of the rectangular ring
        graphics.drawLine(0, 0, 500, 100);
        
        
        // now do the same, but with ellipses
        graphics.setColor(Color.RED);
        Path2D.Double path2 = new Path2D.Double(Path2D.WIND_EVEN_ODD);
        
        // outer circle
        path2.append(new Ellipse2D.Double(500, 400, 100, 150), false);
        
        // Add another circle in center of larger circle
        path2.append(new Ellipse2D.Double(525, 425, 50, 100), false);
        
        graphics.fill(path2);
        
        // draw a line to the center of the elliptical ring
        graphics.drawLine(0, 0, 550, 475);
        
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        // fill a circular ellipse via the path iterator; see that outline is approximately circular
        // (will actually be represented as path with four Bezier approximations to quarter circles
        Ellipse2D.Double pathEllipse = new Ellipse2D.Double(500, 500, 200, 200);
        Path2D.Double path = new Path2D.Double();
        path.append(pathEllipse.getPathIterator(new AffineTransform()), false);
        graphics.fill(path);
        
        graphics.setColor(Color.BLACK);
        
        // draw a line to the center of the circle
        graphics.drawLine(0, 0, 600, 600);
        */
        
    }
    
    
    
    private void testGraphics2D(Graphics2D graphics)
    {
        // do some stuff with Graphics2D and shapes
        
        // save current transform to restore it
        AffineTransform initialTransform = graphics.getTransform();
        
        // draw some stuff in default black and stroke width
        //graphics.setColor(Color.BLACK);
        //graphics.setStroke(new BasicStroke(5));
        
        
        
        double radius = 57.5;
        RealPoint center = new RealPoint(800.3, 627.12, 0);
        
        Line2D line = new Line2D.Double(0, 0, center.x, center.y);
        graphics.draw(line);
        
        Ellipse2D ellipse = new Ellipse2D.Double(center.x-radius, center.y-radius, 2*radius, 2*radius);
        graphics.draw(ellipse);
        
        ellipse = new Ellipse2D.Double(center.x-radius, center.y-radius, 2*radius, 2.1*radius);
        graphics.draw(ellipse);
        ellipse = new Ellipse2D.Double(center.x-radius, center.y-radius, 2*radius, 2.2*radius);
        graphics.draw(ellipse);
        ellipse = new Ellipse2D.Double(center.x-radius, center.y-radius, 2*radius, 2.4*radius);
        graphics.draw(ellipse);
        ellipse = new Ellipse2D.Double(center.x-radius, center.y-radius, 2*radius, 2.8*radius);
        graphics.draw(ellipse);
        
        
        graphics.setColor(Color.RED);
        radius = 50;
        center = new RealPoint(200, 400, 0);
        
        Arc2D arc = new Arc2D.Double(center.x-radius, center.y-radius, 2*radius, 2*radius, 0, 30, Arc2D.OPEN);
        graphics.draw(arc);
        arc = new Arc2D.Double(center.x-radius, center.y-radius, 2*radius, 2.1*radius, 0, 40, Arc2D.OPEN);
        graphics.draw(arc);
        arc = new Arc2D.Double(center.x-radius, center.y-radius, 2*radius, 2.2*radius, 0, 50, Arc2D.OPEN);
        graphics.draw(arc);
        arc = new Arc2D.Double(center.x-radius, center.y-radius, 2*radius, 2.4*radius, 0, 60, Arc2D.OPEN);
        graphics.draw(arc);
        arc = new Arc2D.Double(center.x-radius, center.y-radius, 2*radius, 2.8*radius, 0, 70, Arc2D.OPEN);
        graphics.draw(arc);
        
        
        graphics.setColor(Color.BLUE);
        Rectangle2D rect = new Rectangle2D.Double(45.5, 87.99, 110, 210);
        graphics.draw(rect);
        rect = new Rectangle2D.Double(1.1*45.5, 87.99, 110, 210);
        graphics.draw(rect);
        rect = new Rectangle2D.Double(1.2*45.5, 87.99, 110, 210);
        graphics.draw(rect);
        rect = new Rectangle2D.Double(1.4*45.5, 87.99, 110, 210);
        graphics.draw(rect);
        rect = new Rectangle2D.Double(1.6*45.5, 87.99, 110, 210);
        graphics.draw(rect);
        
        graphics.setColor(Color.GREEN);
        RoundRectangle2D roundRect = new RoundRectangle2D.Double(211.3, 400, 110, 210, 10, 20);
        graphics.draw(roundRect);
        roundRect = new RoundRectangle2D.Double(211.3, 400, 110, 210, 20, 30);
        graphics.draw(roundRect);
        roundRect = new RoundRectangle2D.Double(211.3, 400, 110, 210, 30, 40);
        graphics.draw(roundRect);
        roundRect = new RoundRectangle2D.Double(211.3, 400, 110, 210, 40, 60);
        graphics.draw(roundRect);
        roundRect = new RoundRectangle2D.Double(211.3, 400, 110, 210, 60, 80);
        graphics.draw(roundRect);
        
        
        graphics.setColor(Color.CYAN);
        float[] cp3 = {500, 1600, 500, 1300, 700, 1800, 800, 1200};
        CubicCurve2D cubicSpline = new CubicCurve2D.Float(cp3[0], cp3[1], cp3[2], cp3[3], cp3[4], cp3[5], cp3[6], cp3[7]);
        graphics.draw(cubicSpline);
        
        graphics.setColor(Color.MAGENTA);
        double[] cp2 = {500, 1600, 600, 1400, 800, 1200};
        QuadCurve2D quadraticSpline = new QuadCurve2D.Double(cp2[0], cp2[1], cp2[2], cp2[3], cp2[4], cp2[5]);
        graphics.draw(quadraticSpline);
        
        
        // now create a Path for use with the general draw(Shape) method
        graphics.setColor(Color.BLACK);
        Path2D.Double path1 = new Path2D.Double();
        // Draw square at 1, 1 calling closePath() to draw last segment of square
        path1.moveTo(500.0, 100.0);
        path1.lineTo(600.0, 100.0);
        path1.lineTo(600.0, 200.0);
        path1.lineTo(500.0, 200.0);
        path1.closePath();
        
        // Add another square in center of larger square
        path1.append(new Rectangle.Double(525, 125, 50, 50), false);
        
        graphics.draw(path1);
        
        
        // do an ellipse via the path iterator
        Ellipse2D.Double pathEllipse = new Ellipse2D.Double(500, 500, 200, 200);
        Path2D.Double path = new Path2D.Double();
        path.append(pathEllipse.getPathIterator(new AffineTransform()), false);
        graphics.draw(path);
        
        
        // do a couple of Bezier curves
        double x1 = 200;
        double y1 = 800;
        double x2 = 600;
        double y2 = 1000;
        double ctrlx1 = 400;
        double ctrly1 = 700;
        double ctrlx2 = 450;
        double ctrly2 = 1100;
        CubicCurve2D.Double cubicBezier = new CubicCurve2D.Double(x1, y1, ctrlx1, ctrly1, ctrlx2, ctrly2, x2, y2);
        QuadCurve2D.Double quadraticBezier = new QuadCurve2D.Double(x1, y1, ctrlx1, ctrly1, x2, y2);
        
        graphics.setColor(Color.BLUE);
        graphics.draw(cubicBezier);
        
        // draw in control polygon
        graphics.drawLine((int)x1, (int)y1,(int)ctrlx1, (int)ctrly1);
        graphics.drawLine((int)ctrlx1, (int)ctrly1,(int)ctrlx2, (int)ctrly2);
        graphics.drawLine((int)ctrlx2, (int)ctrly2,(int)x2, (int)y2);
        
        graphics.setColor(Color.RED);
        graphics.draw(quadraticBezier);
     
        // draw in control polygon
        graphics.drawLine((int)x1, (int)y1,(int)ctrlx1, (int)ctrly1);
        graphics.drawLine((int)ctrlx1, (int)ctrly1,(int)x2, (int)y2);
        
        
        // do a cubic Bezier curve with a loop
         x1 = 800;
         y1 = 800;
         x2 = 1000;
         y2 = 1000;
         ctrlx1 = 1200;
         ctrly1 = 700;
         ctrlx2 = 700;
         ctrly2 = 600;
        cubicBezier = new CubicCurve2D.Double(x1, y1, ctrlx1, ctrly1, ctrlx2, ctrly2, x2, y2);
        
        graphics.setColor(Color.BLUE);
        graphics.draw(cubicBezier);
        
        // draw in control polygon
        graphics.drawLine((int)x1, (int)y1,(int)ctrlx1, (int)ctrly1);
        graphics.drawLine((int)ctrlx1, (int)ctrly1,(int)ctrlx2, (int)ctrly2);
        graphics.drawLine((int)ctrlx2, (int)ctrly2,(int)x2, (int)y2);
        
        
        // try a transformation - rotation works
        graphics.rotate(-15 * (Math.PI/180));
        
        graphics.setColor(Color.ORANGE);
        graphics.draw(cubicBezier);
        
        // draw in control polygon
        graphics.drawLine((int)x1, (int)y1,(int)ctrlx1, (int)ctrly1);
        graphics.drawLine((int)ctrlx1, (int)ctrly1,(int)ctrlx2, (int)ctrly2);
        graphics.drawLine((int)ctrlx2, (int)ctrly2,(int)x2, (int)y2);
        
        
        graphics.scale(1.3,1.3);
        
        graphics.setColor(Color.GREEN);
        graphics.draw(cubicBezier);
        
        // draw in control polygon
        graphics.drawLine((int)x1, (int)y1,(int)ctrlx1, (int)ctrly1);
        graphics.drawLine((int)ctrlx1, (int)ctrly1,(int)ctrlx2, (int)ctrly2);
        graphics.drawLine((int)ctrlx2, (int)ctrly2,(int)x2, (int)y2);
        
        
        // draw arcs with different end treatments
        
        graphics.setTransform(initialTransform);
        graphics.setColor(Color.BLUE);
        
        Arc2D.Double arc2D = new Arc2D.Double(600, -100, 100, 200, -30, 120, Arc2D.OPEN);
        graphics.draw(arc2D);
        arc2D = new Arc2D.Double(800, -100, 100, 200, -30, 120, Arc2D.CHORD);
        graphics.draw(arc2D);
        arc2D = new Arc2D.Double(1000, -100, 100, 200, -30, 120, Arc2D.PIE);
        graphics.draw(arc2D);
        
        
    }
    
    
    
    private void newtestTransformationsGraphics(Graphics2D graphics)
    {
        // do some stuff
        graphics.setColor(Color.BLUE);
        graphics.setStroke(new BasicStroke(1));
        
        // save current transform to restore it
        AffineTransform initialTransform = graphics.getTransform();
        
        // Skew: draw a series of increasingly skewed circular arcs and rectangles
        // SHEARING IS BROKEN IN LIBREOFFICE 6.3.5.2!!!! Any transform that includes it 
        // will be incorrectly interpreted in drawing shapes. This is a known bug :-(
        //    https://bugs.documentfoundation.org/show_bug.cgi?id=41542
         
        
        graphics.setColor(Color.BLUE);
        for (int i = 0; i < 3; i++)
        {
            graphics.drawLine(0, 0, 200, 200);
            graphics.drawOval(200, 200, 400, 400);
            //graphics.drawArc(100, 100, 200, 200, 0, 360);
            graphics.drawRect(200, 200, 400, 400);
            graphics.shear(0.2f, 0.3f);
            graphics.setColor(new Color(1.0f/(i+1), 0, 1.0f));
        }
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        // do some stuff with rectanfgles
        graphics.setColor(Color.CYAN);
        
        graphics.drawLine(0, 0, 100, 100);
        graphics.drawRect(100, 100, 200, 200);
        
        graphics.shear(0.2f, 0.1f);
        graphics.drawLine(0, 0, 100, 100);
        graphics.drawRect(100, 100, 200, 200);
        graphics.setTransform(initialTransform);
        
        graphics.rotate(15*Math.PI/180);
        graphics.drawLine(0, 0, 100, 100);
        graphics.drawRect(100, 100, 200, 200);
        graphics.setTransform(initialTransform);
        
        graphics.rotate(-90*Math.PI/180);
        graphics.drawLine(0, 0, 100, 100);
        graphics.drawRect(100, 100, 200, 200);
        graphics.setTransform(initialTransform);
        
        graphics.scale(1.2f, 0.8f);
        graphics.drawLine(0, 0, 100, 100);
        graphics.drawRect(100, 100, 200, 200);
        graphics.setTransform(initialTransform);
        
        // combine scale and rotate and translate
        graphics.translate(100, 300);
        graphics.rotate(-40); // counterclockwise...
        graphics.scale(1.3f, 0.8f);
        graphics.drawLine(0, 0, 100, 100);
        graphics.drawRect(100, 100, 200, 200);
        graphics.setTransform(initialTransform);
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        /*
       
        // combine scale and rotate and translate
        graphics.setColor(Color.BLACK);
        graphics.translate(100, 300);
        graphics.rotate(-40); // counterclockwise...
        graphics.scale(1.3f, 0.8f);
        graphics.drawLine(0, 0, 80, 950);
        graphics.drawRect(80, 880, 620, 140);
        graphics.drawArc(80, 880, 620, 140, -45, 270);
        
        
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        */

    }
    
    
    
    private void testRandom(Graphics2D graphics)
    {
        // do some stuff
        graphics.setColor(randomColor);
        graphics.setStroke(new BasicStroke(3));
        
        // save current transform to restore it
        AffineTransform initialTransform = graphics.getTransform();
        
        //for (int i = 0; i < 2; i++)
        {
                       
            graphics.rotate(randomRotation * (Math.PI/180)); // counterclockwise...
            graphics.shear(randomShearX, randomShearY);
            graphics.scale(randomScaleX, randomScaleY);
            
            graphics.drawLine(0, 0, 100, 100);
            graphics.drawRect(100, 100, 200, 100);
            
            graphics.drawLine(0, 0, 150, 400);
            graphics.drawArc(100, 300, 100, 200, (int)randomStartAngle, (int)randomSweepAngle);
            
            graphics.drawLine(0, 0, 100, 500);
            int[] xPoints = {100, 300, 250};
            int[] yPoints = {500, 450, 600};
            graphics.drawPolygon(xPoints , yPoints , xPoints.length);
            
            Path2D.Double path1 = new Path2D.Double();
            
            graphics.drawLine(0, 0, 500, 100);
            // Draw square at 1, 1 calling closePath() to draw last segment of square
            path1.moveTo(500.0, 100.0);
            path1.lineTo(600.0, 100.0);
            path1.lineTo(600.0, 200.0);
            path1.lineTo(500.0, 200.0);
            path1.closePath();
            
            // Add another square in center of larger square
            path1.append(new Rectangle.Double(525, 125, 50, 50), false);
            
            graphics.draw(path1);
            
            System.out.println("Rotation = " + randomRotation);
            System.out.println("ShearX   = " + randomShearX);
            System.out.println("ShearY   = " + randomShearY);
            System.out.println("ScaleX   = " + randomScaleX);
            System.out.println("ScaleY   = " + randomScaleY);
            
            //graphics.setColor(new Color(1.0f/(i+1), 0, 1.0f));
            
        }
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
    }
    
    
    
    private void testTransformationsGraphics(Graphics2D graphics)
    {
        // do some stuff
        // test line types
        float[] dashArray;
        BasicStroke stroke;
        dashArray = new float[2];
        dashArray[0] = 2; 
        dashArray[1] = 5;
        //stroke = new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 1, dashArray, 0);
        //graphics.setStroke(stroke);
        graphics.setColor(Color.BLUE);
        
        // save current transform to restore it
        AffineTransform initialTransform = graphics.getTransform();
        
  
        
        graphics.drawLine(0, 0, 3*72, 3*72);
        graphics.drawRect(3*72, 3*72, 72, 2*72);
        
        graphics.rotate(-45 * (Math.PI/180)); // counterclockwise...
        
        graphics.drawLine(0, 0, 3*72, 3*72);
        graphics.drawRect(3*72, 3*72, 72, 2*72);
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        // Rotation: draw a series of rotated elliptical arcs
        graphics.setColor(Color.GREEN);
        
        graphics.drawLine(0, 0, 250, 1550);
        //graphics.drawOval(new RectF(100, 1500, 400, 1600));
        graphics.drawArc(100, 1500, 300, 100, 60, 150);
        
        for (int i = 0; i < 4; i++)
        {
            graphics.rotate(-15 * (Math.PI/180)); // counterclockwise...
            graphics.drawLine(0, 0, 250, 1550);
            //graphics.drawOval(new RectF(100, 1500, 400, 1600));
            graphics.drawArc(100, 1500, 300, 100, 60, 150);
        }
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        // Skew: draw a series of increasingly skewed circular arcs and rectangles
        // SHEARING IS BROKEN IN LIBREOFFICE 6.3.5.2!!!! Any transform that includes it 
        // will be incorrectly interpreted in drawing shapes. This is a known bug :-(
        //    https://bugs.documentfoundation.org/show_bug.cgi?id=41542
         
        graphics.setColor(Color.BLUE);
        
        for (int i = 0; i < 4; i++)
        {
            graphics.drawLine(0, 0, 1000, -50);
            graphics.drawOval(1000, -50, 400, 400);
            //graphics.drawArc(100, 100, 200, 200, 0, 360);
            graphics.drawRect(1000, -50, 400, 400);
            graphics.shear(0.2f, 0.0f);
            graphics.setColor(new Color(1.0f/(i+1), 0, 1.0f));
        }
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        graphics.setColor(Color.CYAN);
        for (int i = 0; i < 4; i++)
        {
            graphics.drawLine(0, 0, 100, 600);
            graphics.drawRect(100, 600, 200, 200);
            graphics.shear(0.2f, 0.1f);
        }
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        graphics.setColor(Color.RED);
        // Scaling: do series of scaled circular arcs
        for (int i = 0; i < 4; i++)
        {
            graphics.drawLine(0, 0, 500, 200);
            //graphics.drawCircle(500, 200, 50);
            graphics.drawArc(450, 150, 100, 100, 0, 360);
            graphics.drawRect(450, 150, 100, 100);
            //graphics.drawLine(0, 0, 200, 500);
            //graphics.drawCircle(500, 200, 50);
            //graphics.drawArc(150, 450, 100, 100, 0, 360);
            graphics.scale(1.2f, 1.4f);
        }
        
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        
        
        
        // Translation: do series of translated circular arcs
        graphics.setColor(Color.MAGENTA);
        for (int i = 0; i < 2; i++)
        {
            graphics.drawLine(0, 0, 700, -350);
            //graphics.drawCircle(500, 200, 50);
            graphics.drawArc(600, -450, 200, 200, -135, 315);
            graphics.translate(200, 100);
        }
        
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
       
        // combine scale and rotate and translate
        graphics.setColor(Color.BLACK);
        graphics.translate(100, 300);
        graphics.rotate(-40); // counterclockwise...
        graphics.scale(1.3f, 0.8f);
        graphics.drawLine(0, 0, 80, 950);
        graphics.drawRect(80, 880, 620, 140);
        graphics.drawArc(80, 880, 620, 140, -45, 270);
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
     

    }
    
    
    
    private void testTransformationsText(Graphics2D graphics)
    {
        FontMetrics fm;
        int ascent, descent, width, baseX, baseY;
        String text;
        
        
        // save current transform to restore it
        AffineTransform initialTransform = graphics.getTransform();
        
        graphics.setStroke(new BasicStroke(1));
        graphics.setColor(Color.GREEN);
        graphics.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 14));
        
        baseX = 100;
        baseY = 100;
        
        for (int i = 0; i < 4; i++)
        {
            double scaleFactor = 0.4 + i * 0.3;
            graphics.scale(scaleFactor, scaleFactor);
            
            text = "Some 14-point sans-serif green text at position position (" + baseX + "," + baseY + ") at scale " + (int)Math.round(scaleFactor*100) + "%";
            
            graphics.drawLine(0, 0, baseX, baseY);
            graphics.drawString(text, baseX, baseY);
            
            // use the FontMetrics to draw a box around the text. Unfortunately, since the DXF fonts will in general be different
            // from the corresponding Java fonts, the boxes will be the wrong size in the DXF display.
            fm = graphics.getFontMetrics();
            ascent = fm.getAscent();
            descent = fm.getDescent();
            width = fm.stringWidth(text);
            graphics.drawRect(baseX, baseY - ascent, width, ascent + descent);
            
            graphics.setTransform(initialTransform);
        }
        
        // try some justifications with scaling
        graphics.setFont(new Font(Font.SERIF, Font.PLAIN, 14));
        
        for (int i = 0; i < 2; i++)
        {
            double scaleFactor = 0.7 + i * 0.3;
            graphics.scale(scaleFactor, scaleFactor);
            
            baseX = 200;
            baseY = 300;
            
            text = "Some 14-point serif green text, right aligned at position (" + baseX + "," + baseY + ") at scale " + (int)Math.round(scaleFactor*100) + "%";
            
            graphics.drawLine(0, 0, baseX, baseY);
            
            // use the FontMetrics to align and draw a box around the text. 
            fm = graphics.getFontMetrics();
            ascent = fm.getAscent();
            descent = fm.getDescent();
            width = fm.stringWidth(text);
            
            // adjust base point to get right aligned
            baseX -= width;
            
            
            graphics.drawString(text, baseX, baseY);
            graphics.drawRect(baseX, baseY - ascent, width, ascent + descent);
            
            graphics.setTransform(initialTransform);
            
            
        }
        
        // draw a rotated rectangle and some text
        graphics.setStroke(new BasicStroke(3));
        graphics.setColor(Color.GREEN);
        graphics.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 40));
        
        
        for (int i = 0; i < 4; i ++)
        {
            graphics.rotate(15 * Math.PI/180); // counterclockwise...
            graphics.drawLine(0, 0, 250, 500);
            graphics.drawRect(250, 500, 700, 520);
            graphics.drawString("Some 40-point green text at position 250,500 rotated by " + i*15 + " degrees", 250, 500);
        }
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        graphics.setColor(Color.BLUE);
        // graphics.scale(2, 2);
        graphics.shear(0.3f, 0);
        graphics.drawLine(0, 0, 75, 380);
        graphics.drawRect(75, 380, 700, 520);
        graphics.drawString("Some 40-point blue text skewed at position 80, 450", 80, 450);
        
        // graphics.setMatrix(origMatrix);
        graphics.shear(-0.3f, 0);
        
        graphics.setColor(Color.BLUE);
        graphics.shear(0, 0.8f);
        graphics.drawLine(0, 0, 80, 950);
        graphics.drawRect(80, 880, 700, 1020);
        graphics.drawString("Some 40-point blue text skewed at position 80, 950", 80, 950);
        
        graphics.shear(0, -0.8f);
        
        //restore the original transform
        graphics.setTransform(initialTransform);
        
        
        // both skew and rotate and translate
        graphics.setColor(Color.RED);
        graphics.setFont(new Font(Font.SERIF, Font.PLAIN, 30));
        graphics.translate(100, 300);
        graphics.rotate(-40); // counterclockwise...
        graphics.scale(1.2f, 1.1f);
        graphics.drawLine(0, 0, 80, 950);
        graphics.drawRect(80, 880, 700, 1020);
        graphics.drawString("Some 30-point serif red text at position 80, 950, translated, rotated, scaled", 80, 950);
        
    }
    
    
    
    /**
     * Helper class for some of the grid-related tests
     * 
     * @author jsevy
     *
     */
    public class Vector3D
    {
        float x;
        float y;
        float z;
        
        
        
        public Vector3D(float x, float y, float z)
        {
            this.x = x;
            this.y = y;
            this.z = z;
        }
        
        
        
        public Vector3D()
        {
            this.x = 0;
            this.y = 0;
            this.z = 0;
        }
        
        
        
        public float dotProduct(Vector3D other)
        {
            return this.x * other.x + this.y * other.y + this.z * other.z;
        }
        
        
        
        public float magnitude()
        {
            return (float) Math.sqrt(this.dotProduct(this));
        }
        
        
        
        public Vector3D scalarProduct(float scale)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.x * scale;
            result.y = this.y * scale;
            result.z = this.z * scale;
            
            return result;
        }
        
        
        
        public Vector3D crossProduct(Vector3D other)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.y * other.z - other.y * this.z;
            result.y = this.x * other.z - other.x * this.z;
            result.z = this.x * other.y - other.x * this.y;
            
            return result;
        }
        
        
        
        public Vector3D sum(Vector3D other)
        {
            Vector3D result = new Vector3D();
            
            result.x = this.x + other.x;
            result.y = this.y + other.y;
            result.z = this.z + other.z;
            
            return result;
        }
    }
}
